@@ -1,24 +1,22 @@
+from __future__ import annotations
+
 import asyncio
+import dataclasses
 from abc import ABC, abstractmethod
+from dataclasses import dataclass
 from datetime import datetime, timedelta
 from enum import Enum, IntEnum, unique
 from functools import singledispatch
 from types import TracebackType
 from typing import (
     Any, AsyncContextManager, AsyncIterable, Awaitable, Callable, Dict,
-    FrozenSet, Generator, Iterator, MutableMapping, NamedTuple, Optional, Set,
-    Tuple, Type, TypeVar, Union,
+    Generator, Iterator, Literal, Mapping, Optional, Tuple, Type, TypedDict,
+    TypeVar, Union, overload,
 )
 
-
-try:
-    from typing import TypedDict
-except ImportError:
-    from typing_extensions import TypedDict
-
 import aiormq.abc
 from aiormq.abc import ExceptionType
-from pamqp.common import Arguments
+from pamqp.common import Arguments, FieldValue
 from yarl import URL
 
 from .pool import PoolInstance
@@ -30,12 +28,11 @@ from .tools import (
 TimeoutType = Optional[Union[int, float]]
 
 NoneType = type(None)
-DateType = Union[int, datetime, float, timedelta, None]
+DateType = Optional[Union[int, datetime, float, timedelta]]
 ExchangeParamType = Union["AbstractExchange", str]
 ConsumerTag = str
 
 MILLISECONDS = 1000
-ZERO_TIME = datetime.utcfromtimestamp(0)
 
 
 class SSLOptions(TypedDict, total=False):
@@ -72,7 +69,8 @@ class TransactionState(str, Enum):
     STARTED = "started"
 
 
-class DeclarationResult(NamedTuple):
+@dataclasses.dataclass(frozen=True)
+class DeclarationResult:
     message_count: int
     consumer_count: int
 
@@ -80,11 +78,6 @@ class DeclarationResult(NamedTuple):
 class AbstractTransaction:
     state: TransactionState
 
-    @property
-    @abstractmethod
-    def channel(self) -> "AbstractChannel":
-        raise NotImplementedError
-
     @abstractmethod
     async def select(
         self, timeout: TimeoutType = None,
@@ -114,20 +107,38 @@ class AbstractTransaction:
         raise NotImplementedError
 
 
-HeadersValue = Union[aiormq.abc.FieldValue, bytes]
-HeadersPythonValues = Union[
-    HeadersValue,
-    Set[HeadersValue],
-    Tuple[HeadersValue, ...],
-    FrozenSet[HeadersValue],
-]
-HeadersType = MutableMapping[str, HeadersPythonValues]
+HeadersType = Dict[str, FieldValue]
+
+
+class MessageInfo(TypedDict, total=False):
+    app_id: Optional[str]
+    body_size: int
+    cluster_id: Optional[str]
+    consumer_tag: Optional[str]
+    content_encoding: Optional[str]
+    content_type: Optional[str]
+    correlation_id: Optional[str]
+    delivery_mode: DeliveryMode
+    delivery_tag: Optional[int]
+    exchange: Optional[str]
+    expiration: Optional[DateType]
+    headers: HeadersType
+    message_id: Optional[str]
+    priority: Optional[int]
+    redelivered: Optional[bool]
+    routing_key: Optional[str]
+    reply_to: Optional[str]
+    timestamp: Optional[datetime]
+    type: str
+    user_id: Optional[str]
 
 
 class AbstractMessage(ABC):
+    __slots__ = ()
+
     body: bytes
     body_size: int
-    headers_raw: aiormq.abc.FieldTable
+    headers: HeadersType
     content_type: Optional[str]
     content_encoding: Optional[str]
     delivery_mode: DeliveryMode
@@ -141,17 +152,8 @@ class AbstractMessage(ABC):
     user_id: Optional[str]
     app_id: Optional[str]
 
-    @property
     @abstractmethod
-    def headers(self) -> HeadersType:
-        raise NotImplementedError
-
-    @headers.setter
-    def headers(self, value: HeadersType) -> None:
-        raise NotImplementedError
-
-    @abstractmethod
-    def info(self) -> Dict[str, HeadersValue]:
+    def info(self) -> MessageInfo:
         raise NotImplementedError
 
     @property
@@ -177,6 +179,8 @@ class AbstractMessage(ABC):
 
 
 class AbstractIncomingMessage(AbstractMessage, ABC):
+    __slots__ = ()
+
     cluster_id: Optional[str]
     consumer_tag: Optional["ConsumerTag"]
     delivery_tag: Optional[int]
@@ -211,7 +215,7 @@ class AbstractIncomingMessage(AbstractMessage, ABC):
     async def nack(self, multiple: bool = False, requeue: bool = True) -> None:
         raise NotImplementedError
 
-    def info(self) -> Dict[str, Any]:
+    def info(self) -> MessageInfo:
         raise NotImplementedError
 
     @property
@@ -236,7 +240,9 @@ class AbstractProcessContext(AsyncContextManager):
 
 
 class AbstractQueue:
-    channel: aiormq.abc.AbstractChannel
+    __slots__ = ()
+
+    channel: "AbstractChannel"
     name: str
     durable: bool
     exclusive: bool
@@ -244,7 +250,10 @@ class AbstractQueue:
     arguments: Arguments
     passive: bool
     declaration_result: aiormq.spec.Queue.DeclareOk
-    close_callbacks: CallbackCollection
+    close_callbacks: CallbackCollection[
+        AbstractQueue,
+        [Optional[BaseException]],
+    ]
 
     @abstractmethod
     def __init__(
@@ -257,7 +266,17 @@ class AbstractQueue:
         arguments: Arguments,
         passive: bool = False,
     ):
-        raise NotImplementedError
+        raise NotImplementedError(
+            dict(
+                channel=channel,
+                name=name,
+                durable=durable,
+                exclusive=exclusive,
+                auto_delete=auto_delete,
+                arguments=arguments,
+                passive=passive,
+            ),
+        )
 
     @abstractmethod
     async def declare(
@@ -289,7 +308,7 @@ class AbstractQueue:
     @abstractmethod
     async def consume(
         self,
-        callback: Callable[[AbstractIncomingMessage], Any],
+        callback: Callable[[AbstractIncomingMessage], Awaitable[Any]],
         no_ack: bool = False,
         exclusive: bool = False,
         arguments: Arguments = None,
@@ -306,6 +325,20 @@ class AbstractQueue:
     ) -> aiormq.spec.Basic.CancelOk:
         raise NotImplementedError
 
+    @overload
+    async def get(
+        self, *, no_ack: bool = False,
+        fail: Literal[True] = ..., timeout: TimeoutType = ...,
+    ) -> AbstractIncomingMessage:
+        ...
+
+    @overload
+    async def get(
+        self, *, no_ack: bool = False,
+        fail: Literal[False] = ..., timeout: TimeoutType = ...,
+    ) -> Optional[AbstractIncomingMessage]:
+        ...
+
     @abstractmethod
     async def get(
         self, *, no_ack: bool = False,
@@ -331,14 +364,14 @@ class AbstractQueue:
         raise NotImplementedError
 
 
-class AbstractQueueIterator(AsyncIterable):
+class AbstractQueueIterator(AsyncIterable[AbstractIncomingMessage]):
     _amqp_queue: AbstractQueue
     _queue: asyncio.Queue
     _consumer_tag: ConsumerTag
     _consume_kwargs: Dict[str, Any]
 
     @abstractmethod
-    def close(self, *_: Any) -> Awaitable[Any]:
+    def close(self) -> Awaitable[Any]:
         raise NotImplementedError
 
     @abstractmethod
@@ -377,7 +410,7 @@ class AbstractExchange(ABC):
     @abstractmethod
     def __init__(
         self,
-        channel: aiormq.abc.AbstractChannel,
+        channel: "AbstractChannel",
         name: str,
         type: Union[ExchangeType, str] = ExchangeType.DIRECT,
         *,
@@ -435,7 +468,8 @@ class AbstractExchange(ABC):
         raise NotImplementedError
 
 
-class UnderlayChannel(NamedTuple):
+@dataclasses.dataclass(frozen=True)
+class UnderlayChannel:
     channel: aiormq.abc.AbstractChannel
     close_callback: OneShotCallback
 
@@ -467,7 +501,7 @@ class UnderlayChannel(NamedTuple):
 
         self.channel.closing.remove_done_callback(self.close_callback)
         self.channel.connection.closing.remove_done_callback(
-            self.close_callback
+            self.close_callback,
         )
 
 
@@ -475,8 +509,14 @@ class AbstractChannel(PoolInstance, ABC):
     QUEUE_CLASS: Type[AbstractQueue]
     EXCHANGE_CLASS: Type[AbstractExchange]
 
-    close_callbacks: CallbackCollection
-    return_callbacks: CallbackCollection
+    close_callbacks: CallbackCollection[
+        AbstractChannel,
+        [Optional[BaseException]],
+    ]
+    return_callbacks: CallbackCollection[
+        AbstractChannel,
+        [AbstractIncomingMessage],
+    ]
     default_exchange: AbstractExchange
 
     publisher_confirms: bool
@@ -495,9 +535,12 @@ class AbstractChannel(PoolInstance, ABC):
     def close(self, exc: Optional[ExceptionType] = None) -> Awaitable[None]:
         raise NotImplementedError
 
-    @property
     @abstractmethod
-    def channel(self) -> aiormq.abc.AbstractChannel:
+    def closed(self) -> Awaitable[Literal[True]]:
+        raise NotImplementedError
+
+    @abstractmethod
+    async def get_underlay_channel(self) -> aiormq.abc.AbstractChannel:
         raise NotImplementedError
 
     @property
@@ -612,7 +655,8 @@ class AbstractChannel(PoolInstance, ABC):
         raise NotImplementedError
 
 
-class UnderlayConnection(NamedTuple):
+@dataclasses.dataclass(frozen=True)
+class UnderlayConnection:
     connection: aiormq.abc.AbstractConnection
     close_callback: OneShotCallback
 
@@ -663,10 +707,32 @@ class UnderlayConnection(NamedTuple):
             await self.close_callback.wait()
 
 
+@dataclass
+class ConnectionParameter:
+    name: str
+    parser: Callable[[str], Any]
+    default: Optional[str] = None
+    is_kwarg: bool = False
+
+    def parse(self, value: Optional[str]) -> Any:
+        if value is None:
+            return self.default
+        try:
+            return self.parser(value)
+        except ValueError:
+            return self.default
+
+
 class AbstractConnection(PoolInstance, ABC):
-    close_callbacks: CallbackCollection
+    PARAMETERS: Tuple[ConnectionParameter, ...]
+
+    close_callbacks: CallbackCollection[
+        AbstractConnection,
+        [Optional[BaseException]],
+    ]
     connected: asyncio.Event
     transport: Optional[UnderlayConnection]
+    kwargs: Mapping[str, Any]
 
     @abstractmethod
     def __init__(
@@ -686,6 +752,10 @@ class AbstractConnection(PoolInstance, ABC):
     async def close(self, exc: ExceptionType = asyncio.CancelledError) -> None:
         raise NotImplementedError
 
+    @abstractmethod
+    def closed(self) -> Awaitable[Literal[True]]:
+        raise NotImplementedError
+
     @abstractmethod
     async def connect(self, timeout: TimeoutType = None) -> None:
         raise NotImplementedError
@@ -725,8 +795,10 @@ class AbstractConnection(PoolInstance, ABC):
 
 
 class AbstractRobustQueue(AbstractQueue):
+    __slots__ = ()
+
     @abstractmethod
-    def restore(self, channel: aiormq.abc.AbstractChannel) -> Awaitable[None]:
+    def restore(self) -> Awaitable[None]:
         raise NotImplementedError
 
     @abstractmethod
@@ -757,7 +829,7 @@ class AbstractRobustQueue(AbstractQueue):
 
 class AbstractRobustExchange(AbstractExchange):
     @abstractmethod
-    def restore(self, channel: aiormq.abc.AbstractChannel) -> Awaitable[None]:
+    def restore(self) -> Awaitable[None]:
         raise NotImplementedError
 
     @abstractmethod
@@ -774,14 +846,14 @@ class AbstractRobustExchange(AbstractExchange):
 
 
 class AbstractRobustChannel(AbstractChannel):
-    reopen_callbacks: CallbackCollection
+    reopen_callbacks: CallbackCollection[AbstractRobustChannel, []]
 
     @abstractmethod
     def reopen(self) -> Awaitable[None]:
         raise NotImplementedError
 
     @abstractmethod
-    async def restore(self, connection: aiormq.abc.AbstractConnection) -> None:
+    async def restore(self) -> None:
         raise NotImplementedError
 
     @abstractmethod
@@ -817,7 +889,7 @@ class AbstractRobustChannel(AbstractChannel):
 
 
 class AbstractRobustConnection(AbstractConnection):
-    reconnect_callbacks: CallbackCollection
+    reconnect_callbacks: CallbackCollection[AbstractRobustConnection, []]
 
     @property
     @abstractmethod
@@ -839,10 +911,10 @@ class AbstractRobustConnection(AbstractConnection):
 
 
 ChannelCloseCallback = Callable[
-    [AbstractChannel, Optional[BaseException]], Any,
+    [Optional[AbstractChannel], Optional[BaseException]], Any,
 ]
 ConnectionCloseCallback = Callable[
-    [AbstractConnection, Optional[BaseException]], Any,
+    [Optional[AbstractConnection], Optional[BaseException]], Any,
 ]
 ConnectionType = TypeVar("ConnectionType", bound=AbstractConnection)
 
@@ -883,21 +955,22 @@ __all__ = (
     "CallbackType",
     "ChannelCloseCallback",
     "ConnectionCloseCallback",
+    "ConnectionParameter",
     "ConsumerTag",
     "DateType",
     "DeclarationResult",
     "DeliveryMode",
     "ExchangeParamType",
     "ExchangeType",
-    "HeadersPythonValues",
+    "FieldValue",
     "HeadersType",
-    "HeadersValue",
     "MILLISECONDS",
+    "MessageInfo",
+    "NoneType",
     "SSLOptions",
     "TimeoutType",
     "TransactionState",
     "UnderlayChannel",
     "UnderlayConnection",
-    "ZERO_TIME",
     "get_exchange_name",
 )
