@@ -1,5 +1,9 @@
 import pickle
-from typing import Any, Callable
+from typing import Any, Awaitable, Callable, TypeVar
+
+
+T = TypeVar("T")
+CallbackType = Callable[..., Awaitable[T]]
 
 
 class Method:
@@ -30,6 +34,8 @@ class Proxy:
 
 
 class Base:
+    __slots__ = ()
+
     SERIALIZER = pickle
     CONTENT_TYPE = "application/python-pickle"
 
