@@ -1,21 +1,28 @@
 import asyncio
+from asyncio import Future
 from functools import partial
 from types import TracebackType
-from typing import Any, Callable, Optional, Type
+from typing import (
+    Any, Awaitable, Callable, Literal, Optional, Type, cast, overload,
+)
 
 import aiormq
 from aiormq.abc import DeliveredMessage
+try:
+    ExceptionGroup
+except NameError:
+    from exceptiongroup import ExceptionGroup
 from pamqp.common import Arguments
 
 from .abc import (
-    AbstractIncomingMessage, AbstractQueue, AbstractQueueIterator, ConsumerTag,
-    TimeoutType, get_exchange_name,
+    AbstractChannel, AbstractIncomingMessage, AbstractQueue,
+    AbstractQueueIterator, ConsumerTag, TimeoutType, get_exchange_name,
 )
 from .exceptions import QueueEmpty
 from .exchange import ExchangeParamType
 from .log import get_logger
 from .message import IncomingMessage
-from .tools import CallbackCollection, create_task
+from .tools import CallbackCollection, create_task, ensure_awaitable
 
 
 log = get_logger(__name__)
@@ -24,7 +31,7 @@ log = get_logger(__name__)
 async def consumer(
     callback: Callable[[AbstractIncomingMessage], Any],
     msg: DeliveredMessage, *,
-    no_ack: bool
+    no_ack: bool,
 ) -> Any:
     message = IncomingMessage(msg, no_ack=no_ack)
     return await create_task(callback, message)
@@ -33,9 +40,23 @@ async def consumer(
 class Queue(AbstractQueue):
     """ AMQP queue abstraction """
 
+    __slots__ = (
+        "__weakref__",
+        "__get_lock",
+        "close_callbacks",
+        "channel",
+        "name",
+        "durable",
+        "exclusive",
+        "auto_delete",
+        "arguments",
+        "passive",
+        "declaration_result",
+    )
+
     def __init__(
         self,
-        channel: aiormq.abc.AbstractChannel,
+        channel: AbstractChannel,
         name: Optional[str],
         durable: bool,
         exclusive: bool,
@@ -71,11 +92,11 @@ class Queue(AbstractQueue):
         """ Declare queue.
 
         :param timeout: execution timeout
-        :param passive: Only check to see if the queue exists.
         :return: :class:`None`
         """
         log.debug("Declaring queue: %r", self)
-        self.declaration_result = await self.channel.queue_declare(
+        channel = await self.channel.get_underlay_channel()
+        self.declaration_result = await channel.queue_declare(
             queue=self.name,
             durable=self.durable,
             exclusive=self.exclusive,
@@ -97,7 +118,7 @@ class Queue(AbstractQueue):
         routing_key: Optional[str] = None,
         *,
         arguments: Arguments = None,
-        timeout: TimeoutType = None
+        timeout: TimeoutType = None,
     ) -> aiormq.spec.Queue.BindOk:
 
         """ A binding is a relationship between an exchange and a queue.
@@ -128,7 +149,8 @@ class Queue(AbstractQueue):
             arguments,
         )
 
-        return await self.channel.queue_bind(
+        channel = await self.channel.get_underlay_channel()
+        return await channel.queue_bind(
             self.name,
             exchange=get_exchange_name(exchange),
             routing_key=routing_key,
@@ -166,7 +188,8 @@ class Queue(AbstractQueue):
             arguments,
         )
 
-        return await self.channel.queue_unbind(
+        channel = await self.channel.get_underlay_channel()
+        return await channel.queue_unbind(
             queue=self.name,
             exchange=get_exchange_name(exchange),
             routing_key=routing_key,
@@ -176,7 +199,7 @@ class Queue(AbstractQueue):
 
     async def consume(
         self,
-        callback: Callable[[AbstractIncomingMessage], Any],
+        callback: Callable[[AbstractIncomingMessage], Awaitable[Any]],
         no_ack: bool = False,
         exclusive: bool = False,
         arguments: Arguments = None,
@@ -188,7 +211,7 @@ class Queue(AbstractQueue):
 
         :param timeout: :class:`asyncio.TimeoutError` will be raises when the
                         Future was not finished after this time.
-        :param callback: Consuming callback. Could be a coroutine.
+        :param callback: Consuming callback. Should be a coroutine function.
         :param no_ack:
             if :class:`True` you don't need to call
             :func:`aio_pika.message.IncomingMessage.ack`
@@ -207,8 +230,10 @@ class Queue(AbstractQueue):
         """
 
         log.debug("Start to consuming queue: %r", self)
+        callback = ensure_awaitable(callback)
 
-        consume_result = await self.channel.basic_consume(
+        channel = await self.channel.get_underlay_channel()
+        consume_result = await channel.basic_consume(
             queue=self.name,
             consumer_callback=partial(
                 consumer,
@@ -252,13 +277,28 @@ class Queue(AbstractQueue):
         :return: Basic.CancelOk when operation completed successfully
         """
 
-        return await self.channel.basic_cancel(
+        channel = await self.channel.get_underlay_channel()
+        return await channel.basic_cancel(
             consumer_tag=consumer_tag, nowait=nowait, timeout=timeout,
         )
 
+    @overload
     async def get(
         self, *, no_ack: bool = False,
-        fail: bool = True, timeout: TimeoutType = 5
+        fail: Literal[True] = ..., timeout: TimeoutType = ...,
+    ) -> IncomingMessage:
+        ...
+
+    @overload
+    async def get(
+        self, *, no_ack: bool = False,
+        fail: Literal[False] = ..., timeout: TimeoutType = ...,
+    ) -> Optional[IncomingMessage]:
+        ...
+
+    async def get(
+        self, *, no_ack: bool = False,
+        fail: bool = True, timeout: TimeoutType = 5,
     ) -> Optional[IncomingMessage]:
 
         """ Get message from the queue.
@@ -271,7 +311,8 @@ class Queue(AbstractQueue):
         :return: :class:`aio_pika.message.IncomingMessage`
         """
 
-        msg: DeliveredMessage = await self.channel.basic_get(
+        channel = await self.channel.get_underlay_channel()
+        msg: DeliveredMessage = await channel.basic_get(
             self.name, no_ack=no_ack, timeout=timeout,
         )
 
@@ -294,13 +335,14 @@ class Queue(AbstractQueue):
 
         log.info("Purging queue: %r", self)
 
-        return await self.channel.queue_purge(
+        channel = await self.channel.get_underlay_channel()
+        return await channel.queue_purge(
             self.name, nowait=no_wait, timeout=timeout,
         )
 
     async def delete(
         self, *, if_unused: bool = True,
-        if_empty: bool = True, timeout: TimeoutType = None
+        if_empty: bool = True, timeout: TimeoutType = None,
     ) -> aiormq.spec.Queue.DeleteOk:
 
         """ Delete the queue.
@@ -313,7 +355,8 @@ class Queue(AbstractQueue):
 
         log.info("Deleting %r", self)
 
-        return await self.channel.queue_delete(
+        channel = await self.channel.get_underlay_channel()
+        return await channel.queue_delete(
             self.name,
             if_unused=if_unused,
             if_empty=if_empty,
@@ -377,7 +420,24 @@ class QueueIterator(AbstractQueueIterator):
     def consumer_tag(self) -> Optional[ConsumerTag]:
         return getattr(self, "_consumer_tag", None)
 
-    async def close(self, *_: Any) -> Any:
+    async def close(self) -> None:
+        await self._on_close(self._amqp_queue, None)
+        if not self._closed.done():
+            self._closed.set_result(True)
+
+    async def _set_closed(
+        self,
+        _channel: Optional[AbstractQueue],
+        exc: Optional[BaseException]
+    ) -> None:
+        if not self._closed.done():
+            self._closed.set_result(True)
+
+    async def _on_close(
+        self,
+        _channel: Optional[AbstractQueue],
+        _exc: Optional[BaseException]
+    ) -> None:
         log.debug("Cancelling queue iterator %r", self)
 
         if not hasattr(self, "_consumer_tag"):
@@ -392,19 +452,47 @@ class QueueIterator(AbstractQueueIterator):
         consumer_tag = self._consumer_tag
         del self._consumer_tag
 
-        self._amqp_queue.close_callbacks.remove(self.close)
+        self._amqp_queue.close_callbacks.discard(self._on_close)
         await self._amqp_queue.cancel(consumer_tag)
 
         log.debug("Queue iterator %r closed", self)
 
-        # Reject all messages
-        msg: Optional[IncomingMessage] = None
-        try:
-            while True:
-                msg = self._queue.get_nowait()
-        except asyncio.QueueEmpty:
-            if msg is not None and not self._amqp_queue.channel.is_closed:
-                await msg.nack(requeue=True, multiple=True)
+        if self._queue.empty():
+            return
+
+        exceptions = []
+        # Reject all messages that have been received and in the buffer/cache.
+        while not self._queue.empty():
+            msg = self._queue.get_nowait()
+
+            if self._amqp_queue.channel.is_closed:
+                log.warning(
+                    "Message %r lost when queue iterator %r channel closed",
+                    msg,
+                    self,
+                )
+            elif self._consume_kwargs.get("no_ack", False):
+                log.warning(
+                    "Message %r lost for consumer with no_ack %r",
+                    msg,
+                    self,
+                )
+            else:
+                try:
+                    await msg.nack(requeue=True, multiple=False)
+                except Exception as e:
+                    log.warning(
+                        "Failed to nack message %r",
+                        msg,
+                        exc_info=e,
+                    )
+                    exceptions.append(e)
+
+        if exceptions:
+            raise ExceptionGroup(
+                "Unable to nack all messages",
+                exceptions,
+            )
 
     def __str__(self) -> str:
         return f"queue[{self._amqp_queue}](...)"
@@ -418,18 +506,34 @@ class QueueIterator(AbstractQueueIterator):
 
     def __init__(self, queue: Queue, **kwargs: Any):
         self._consumer_tag: ConsumerTag
-        self._amqp_queue: AbstractQueue = queue
+        self._amqp_queue: Queue = queue
         self._queue = asyncio.Queue()
+        self._closed = asyncio.get_running_loop().create_future()
+        self._message_or_closed = asyncio.Event()
+        self._timeout_event = asyncio.Event()
         self._consume_kwargs = kwargs
 
-        self._amqp_queue.close_callbacks.add(self.close)
+        cast(
+            asyncio.Future, self._amqp_queue.channel.closed()
+        ).add_done_callback(self._propagate_closed)
+        self._closed.add_done_callback(self._propagate_closed)
+
+        self._amqp_queue.close_callbacks.add(self._on_close, weak=True)
+        self._amqp_queue.close_callbacks.add(
+            self._set_closed,
+            weak=True
+        )
+
+    def _propagate_closed(self, _: Future) -> None:
+        self._message_or_closed.set()
 
     async def on_message(self, message: AbstractIncomingMessage) -> None:
         await self._queue.put(message)
+        self._message_or_closed.set()
 
     async def consume(self) -> None:
         self._consumer_tag = await self._amqp_queue.consume(
-            self.on_message, **self._consume_kwargs
+            self.on_message, **self._consume_kwargs,
         )
 
     def __aiter__(self) -> "AbstractQueueIterator":
@@ -446,27 +550,70 @@ class QueueIterator(AbstractQueueIterator):
         exc_val: Optional[BaseException],
         exc_tb: Optional[TracebackType],
     ) -> None:
-        await self.close()
+        if hasattr(self, "__closing"):
+            try:
+                await self.__closing
+            finally:
+                del self.__closing
+        else:
+            await self.close()
+
+    async def __anext__(self) -> AbstractIncomingMessage:
+        if self._closed.done():
+            raise StopAsyncIteration
 
-    async def __anext__(self) -> IncomingMessage:
         if not hasattr(self, "_consumer_tag"):
             await self.consume()
-        try:
-            return await asyncio.wait_for(
-                self._queue.get(),
-                timeout=self._consume_kwargs.get("timeout"),
-            )
-        except asyncio.CancelledError:
-            timeout = self._consume_kwargs.get(
-                "timeout",
-                self.DEFAULT_CLOSE_TIMEOUT,
-            )
-            log.info(
-                "%r closing with timeout %d seconds",
-                self, timeout,
-            )
-            await asyncio.wait_for(self.close(), timeout=timeout)
-            raise
+
+        timeout: Optional[float] = self._consume_kwargs.get("timeout")
+
+        if not self._message_or_closed.is_set():
+            coroutine: Awaitable[Any] = self._message_or_closed.wait()
+            if timeout is not None and timeout > 0:
+                coroutine = asyncio.wait_for(coroutine, timeout=timeout)
+
+            try:
+                await coroutine
+            except (asyncio.TimeoutError, asyncio.CancelledError):
+                if timeout is not None:
+                    timeout = (
+                        timeout
+                        if timeout > 0
+                        else self.DEFAULT_CLOSE_TIMEOUT
+                    )
+                    log.info(
+                        "%r closing with timeout %d seconds",
+                        self, timeout,
+                    )
+
+                task = asyncio.create_task(self.close())
+                coroutine = task
+                if timeout is not None:
+                    coroutine = asyncio.wait_for(
+                        asyncio.shield(coroutine),
+                        timeout=timeout,
+                    )
+
+                try:
+                    await coroutine
+                except asyncio.TimeoutError:
+                    self.__closing = task
+
+                raise
+
+        if self._queue.empty():
+            raise StopAsyncIteration
+
+        msg = self._queue.get_nowait()
+
+        if (
+            self._queue.empty()
+            and not self._amqp_queue.channel.is_closed
+            and not self._closed.done()
+        ):
+            self._message_or_closed.clear()
+
+        return msg
 
 
 __all__ = ("Queue", "QueueIterator", "ConsumerTag")
