@@ -1,4 +1,5 @@
 import asyncio
+import warnings
 from collections import defaultdict
 from itertools import chain
 from typing import Any, DefaultDict, Dict, Optional, Set, Type, Union
@@ -7,8 +8,8 @@ from warnings import warn
 import aiormq
 
 from .abc import (
-    AbstractRobustChannel, AbstractRobustExchange, AbstractRobustQueue,
-    TimeoutType,
+    AbstractConnection, AbstractRobustChannel, AbstractRobustExchange,
+    AbstractRobustQueue, TimeoutType,
 )
 from .channel import Channel
 from .exchange import Exchange, ExchangeType
@@ -28,13 +29,15 @@ class RobustChannel(Channel, AbstractRobustChannel):    # type: ignore
     QUEUE_CLASS: Type[Queue] = RobustQueue
     EXCHANGE_CLASS: Type[Exchange] = RobustExchange
 
+    RESTORE_RETRY_DELAY: int = 2
+
     _exchanges: DefaultDict[str, Set[AbstractRobustExchange]]
     _queues: DefaultDict[str, Set[RobustQueue]]
     default_exchange: RobustExchange
 
     def __init__(
         self,
-        connection: aiormq.abc.AbstractConnection,
+        connection: AbstractConnection,
         channel_number: Optional[int] = None,
         publisher_confirms: bool = True,
         on_return_raises: bool = False,
@@ -62,46 +65,97 @@ class RobustChannel(Channel, AbstractRobustChannel):    # type: ignore
         self._prefetch_count: int = 0
         self._prefetch_size: int = 0
         self._global_qos: bool = False
-        self.reopen_callbacks: CallbackCollection = CallbackCollection(self)
-        self.close_callbacks.add(self.__close_callback)
+        self.reopen_callbacks = CallbackCollection(self)
         self.__restore_lock = asyncio.Lock()
+        self.__restored = asyncio.Event()
 
-    async def __close_callback(self, *_: Any) -> None:
-        if self._closed or self._connection.is_closed:
-            self.close_callbacks.discard(self.__close_callback)
-            return
+        self.close_callbacks.remove(self._set_closed_callback)
 
-        await self.reopen()
+    async def ready(self) -> None:
+        await self._connection.ready()
+        await self.__restored.wait()
+
+    async def get_underlay_channel(self) -> aiormq.abc.AbstractChannel:
+        await self._connection.ready()
+        return await super().get_underlay_channel()
+
+    async def restore(self, channel: Any = None) -> None:
+        if channel is not None:
+            warnings.warn(
+                "Channel argument will be ignored because you "
+                "don't need to pass this anymore.",
+                DeprecationWarning,
+            )
 
-    async def restore(self, connection: aiormq.abc.AbstractConnection) -> None:
         async with self.__restore_lock:
-            self._connection = connection
-            await self.reopen()
+            if self.__restored.is_set():
+                return
 
-    async def _open(self) -> None:
-        await super()._open()
+            await self.reopen()
+            self.__restored.set()
+
+    async def _on_close(
+        self,
+        closing: asyncio.Future
+    ) -> Optional[BaseException]:
+        exc = await super()._on_close(closing)
+
+        if isinstance(exc, asyncio.CancelledError):
+            # This happens only if the channel is forced to close from the
+            # outside, for example, if the connection is closed.
+            # Of course, here you need to exit from this function
+            # as soon as possible and to avoid a recovery attempt.
+            self.__restored.clear()
+            if not self._closed.done():
+                self._closed.set_result(True)
+            return exc
+
+        in_restore_state = not self.__restored.is_set()
+        self.__restored.clear()
+
+        if self._closed.done() or in_restore_state:
+            return exc
+
+        await self.restore()
+
+        return exc
+
+    async def close(
+        self,
+        exc: Optional[aiormq.abc.ExceptionType] = None,
+    ) -> None:
+        # Avoid recovery when channel is explicitely closed using this method
+        self.__restored.clear()
+        await super().close(exc)
+
+    async def reopen(self) -> None:
+        await super().reopen()
         await self.reopen_callbacks()
 
-    async def _on_open(self, channel: aiormq.abc.AbstractChannel) -> None:
+    async def _on_open(self) -> None:
         if not hasattr(self, "default_exchange"):
-            await super()._on_open(channel)
+            await super()._on_open()
 
         exchanges = tuple(chain(*self._exchanges.values()))
         queues = tuple(chain(*self._queues.values()))
+        channel = await self.get_underlay_channel()
 
         await channel.basic_qos(
             prefetch_count=self._prefetch_count,
             prefetch_size=self._prefetch_size,
+            global_=self._global_qos,
         )
 
         for exchange in exchanges:
-            await exchange.restore(channel)
+            await exchange.restore()
 
         for queue in queues:
-            await queue.restore(channel)
+            await queue.restore()
 
         if hasattr(self, "default_exchange"):
-            self.default_exchange.channel = channel
+            self.default_exchange.channel = self
+
+        self.__restored.set()
 
     async def set_qos(
         self,
@@ -115,7 +169,7 @@ class RobustChannel(Channel, AbstractRobustChannel):    # type: ignore
             warn('Use "global_" instead of "all_channels"', DeprecationWarning)
             global_ = all_channels
 
-        await self._connection.ready()
+        await self.ready()
 
         self._prefetch_count = prefetch_count
         self._prefetch_size = prefetch_size
@@ -140,7 +194,12 @@ class RobustChannel(Channel, AbstractRobustChannel):    # type: ignore
         timeout: TimeoutType = None,
         robust: bool = True,
     ) -> AbstractRobustExchange:
-        await self._connection.ready()
+        """
+        :param robust: If True, the exchange will be re-declared during
+        reconnection.
+        Set to False for temporary exchanges that should not be restored.
+        """
+        await self.ready()
         exchange = (
             await super().declare_exchange(
                 name=name,
@@ -167,7 +226,7 @@ class RobustChannel(Channel, AbstractRobustChannel):    # type: ignore
         if_unused: bool = False,
         nowait: bool = False,
     ) -> aiormq.spec.Exchange.DeleteOk:
-        await self._connection.ready()
+        await self.ready()
         result = await super().exchange_delete(
             exchange_name=exchange_name,
             timeout=timeout,
@@ -187,9 +246,14 @@ class RobustChannel(Channel, AbstractRobustChannel):    # type: ignore
         auto_delete: bool = False,
         arguments: Optional[Dict[str, Any]] = None,
         timeout: TimeoutType = None,
-        robust: bool = True
+        robust: bool = True,
     ) -> AbstractRobustQueue:
-        await self._connection.ready()
+        """
+        :param robust: If True, the queue will be re-declared during
+        reconnection.
+        Set to False for temporary queues that should not be restored.
+        """
+        await self.ready()
         queue: RobustQueue = await super().declare_queue(   # type: ignore
             name=name,
             durable=durable,
@@ -211,7 +275,7 @@ class RobustChannel(Channel, AbstractRobustChannel):    # type: ignore
         if_empty: bool = False,
         nowait: bool = False,
     ) -> aiormq.spec.Queue.DeleteOk:
-        await self._connection.ready()
+        await self.ready()
         result = await super().queue_delete(
             queue_name=queue_name,
             timeout=timeout,
