@@ -3,8 +3,8 @@ import logging
 import os
 import time
 import uuid
-from datetime import datetime
-from typing import Awaitable, Callable, Optional
+from datetime import datetime, timezone
+from typing import Callable, Optional, List
 from unittest import mock
 
 import aiormq.exceptions
@@ -15,12 +15,15 @@ from yarl import URL
 import aio_pika
 import aio_pika.exceptions
 from aio_pika import Channel, DeliveryMode, Message
-from aio_pika.abc import AbstractConnection, AbstractIncomingMessage
+from aio_pika.abc import (
+    AbstractConnection, AbstractIncomingMessage, MessageInfo, AbstractQueue,
+)
 from aio_pika.exceptions import (
     DeliveryError, MessageProcessError, ProbableAuthenticationError,
 )
 from aio_pika.exchange import ExchangeType
 from aio_pika.message import ReturnedMessage
+from aio_pika.queue import QueueIterator
 from tests import get_random_name
 
 
@@ -31,7 +34,7 @@ class TestCaseAmqpBase:
     @staticmethod
     def create_channel(
         connection: aio_pika.Connection,
-    ) -> Awaitable[aio_pika.Channel]:
+    ) -> aio_pika.abc.AbstractChannel:
         return connection.channel()
 
     @staticmethod
@@ -46,7 +49,9 @@ class TestCaseAmqpBase:
 
 
 class TestCaseAmqp(TestCaseAmqpBase):
-    async def test_properties(self, loop, connection: aio_pika.Connection):
+    async def test_properties(
+        self, event_loop, connection: aio_pika.Connection,
+    ):
         assert not connection.is_closed
 
     async def test_channel_close(self, connection: aio_pika.Connection):
@@ -55,12 +60,13 @@ class TestCaseAmqp(TestCaseAmqpBase):
         closed = False
 
         def on_close(
-            ch: aio_pika.abc.AbstractChannel,
-            exc: Optional[Exception] = None,
+            ch: Optional[aio_pika.abc.AbstractChannel],
+            exc: Optional[BaseException] = None,
         ):
             nonlocal event, closed
             log.info("Close called")
             closed = True
+            assert ch is not None
             assert ch.is_closed
             event.set()
 
@@ -335,6 +341,34 @@ class TestCaseAmqp(TestCaseAmqpBase):
 
         await queue.unbind(dest_exchange, routing_key)
 
+    async def test_simple_publish_with_closed_channel(
+        self,
+        connection: aio_pika.Connection,
+        declare_exchange: Callable,
+        declare_queue: Callable,
+    ):
+        routing_key = get_random_name()
+
+        channel = await connection.channel(publisher_confirms=False)
+
+        exchange = await declare_exchange(
+            "direct", auto_delete=True, channel=channel,
+        )
+
+        await connection.close()
+
+        body = bytes(shortuuid.uuid(), "utf-8")
+
+        with pytest.raises(aiormq.exceptions.ChannelInvalidStateError):
+            await exchange.publish(
+                Message(
+                    body,
+                    content_type="text/plain",
+                    headers={"foo": "bar"},
+                ),
+                routing_key,
+            )
+
     async def test_incoming_message_info(
         self,
         channel: aio_pika.Channel,
@@ -353,22 +387,22 @@ class TestCaseAmqp(TestCaseAmqpBase):
 
         body = bytes(shortuuid.uuid(), "utf-8")
 
-        info = {
-            "headers": {"foo": "bar"},
-            "content_type": "application/json",
-            "content_encoding": "text",
-            "delivery_mode": DeliveryMode.PERSISTENT.value,
-            "priority": 0,
-            "correlation_id": "1",
-            "reply_to": "test",
-            "expiration": 1.5,
-            "message_id": shortuuid.uuid(),
-            "timestamp": datetime.utcfromtimestamp(int(time.time())),
-            "type": "0",
-            "user_id": "guest",
-            "app_id": "test",
-            "body_size": len(body),
-        }
+        info = MessageInfo(
+            app_id="test",
+            body_size=len(body),
+            content_encoding="text",
+            content_type="application/json",
+            correlation_id="1",
+            delivery_mode=DeliveryMode.PERSISTENT,
+            expiration=1.5,
+            headers={"foo": "bar"},
+            message_id=shortuuid.uuid(),
+            priority=0,
+            reply_to="test",
+            timestamp=datetime.fromtimestamp(int(time.time()), tz=timezone.utc),
+            type="0",
+            user_id="guest",
+        )
 
         msg = Message(
             body=body,
@@ -377,7 +411,7 @@ class TestCaseAmqp(TestCaseAmqpBase):
             content_encoding="text",
             delivery_mode=DeliveryMode.PERSISTENT,
             priority=0,
-            correlation_id=1,
+            correlation_id="1",
             reply_to="test",
             expiration=1.5,
             message_id=info["message_id"],
@@ -702,7 +736,7 @@ class TestCaseAmqp(TestCaseAmqpBase):
 
     async def test_consuming(
         self,
-        loop,
+        event_loop,
         channel: aio_pika.Channel,
         declare_exchange: Callable,
         declare_queue: Callable,
@@ -719,7 +753,7 @@ class TestCaseAmqp(TestCaseAmqpBase):
 
         body = bytes(shortuuid.uuid(), "utf-8")
 
-        f = loop.create_future()
+        f = event_loop.create_future()
 
         async def handle(message):
             await message.ack()
@@ -739,7 +773,7 @@ class TestCaseAmqp(TestCaseAmqpBase):
 
     async def test_consuming_not_coroutine(
         self,
-        loop,
+        event_loop,
         channel: aio_pika.Channel,
         declare_exchange: Callable,
         declare_queue: Callable,
@@ -757,7 +791,7 @@ class TestCaseAmqp(TestCaseAmqpBase):
 
         body = bytes(shortuuid.uuid(), "utf-8")
 
-        f = loop.create_future()
+        f = event_loop.create_future()
 
         async def handle(message):
             await message.ack()
@@ -873,6 +907,12 @@ class TestCaseAmqp(TestCaseAmqpBase):
     async def test_set_qos(self, channel: aio_pika.Channel):
         await channel.set_qos(prefetch_count=1, global_=True)
 
+    async def test_set_qos_deprecated_all_channels(
+        self, channel: aio_pika.Channel,
+    ):
+        with pytest.deprecated_call():
+            await channel.set_qos(prefetch_count=1, all_channels=True)
+
     async def test_exchange_delete(
         self, channel: aio_pika.Channel, declare_exchange,
     ):
@@ -881,7 +921,7 @@ class TestCaseAmqp(TestCaseAmqpBase):
 
     async def test_dlx(
         self,
-        loop,
+        event_loop,
         channel: aio_pika.Channel,
         declare_exchange: Callable,
         declare_queue: Callable,
@@ -891,7 +931,7 @@ class TestCaseAmqp(TestCaseAmqpBase):
         routing_key = "%s_routing_key" % suffix
         dlx_routing_key = "%s_dlx_routing_key" % suffix
 
-        f = loop.create_future()
+        f = event_loop.create_future()
 
         async def dlx_handle(message):
             await message.ack()
@@ -946,7 +986,8 @@ class TestCaseAmqp(TestCaseAmqpBase):
             await f
 
     async def test_expiration(
-        self, channel: aio_pika.Channel, loop, declare_exchange, declare_queue,
+        self, channel: aio_pika.Channel, event_loop,
+        declare_exchange, declare_queue,
     ):
 
         dlx_queue = await declare_queue(
@@ -980,7 +1021,7 @@ class TestCaseAmqp(TestCaseAmqpBase):
             queue.name,
         )
 
-        f = loop.create_future()
+        f = event_loop.create_future()
 
         await dlx_queue.consume(f.set_result, no_ack=True)
 
@@ -1205,7 +1246,9 @@ class TestCaseAmqp(TestCaseAmqpBase):
                 async with channel.transaction():
                     raise ValueError
 
-    async def test_async_for_queue(self, loop, connection, declare_queue):
+    async def test_async_for_queue(
+        self, event_loop, connection, declare_queue,
+    ):
         channel2 = await self.create_channel(connection)
 
         queue = await declare_queue(
@@ -1224,7 +1267,7 @@ class TestCaseAmqp(TestCaseAmqpBase):
                     Message(body=str(i).encode()), routing_key=queue.name,
                 )
 
-        loop.create_task(publisher())
+        event_loop.create_task(publisher())
 
         count = 0
         data = list()
@@ -1240,8 +1283,8 @@ class TestCaseAmqp(TestCaseAmqpBase):
         assert data == list(map(lambda x: str(x).encode(), range(messages)))
 
     async def test_async_for_queue_context(
-        self, loop, connection, declare_queue,
-    ):
+        self, event_loop, connection, declare_queue,
+    ) -> None:
         channel2 = await self.create_channel(connection)
 
         queue = await declare_queue(
@@ -1250,34 +1293,50 @@ class TestCaseAmqp(TestCaseAmqpBase):
             channel=channel2,
         )
 
-        messages = 100
+        messages: asyncio.Queue[bytes] = asyncio.Queue(100)
+        condition = asyncio.Condition()
 
-        async def publisher():
+        async def publisher() -> None:
             channel1 = await self.create_channel(connection)
 
-            for i in range(messages):
+            for i in range(messages.maxsize):
+                body = str(i).encode()
+                await messages.put(body)
                 await channel1.default_exchange.publish(
-                    Message(body=str(i).encode()), routing_key=queue.name,
+                    Message(body=body), routing_key=queue.name,
                 )
 
-        loop.create_task(publisher())
+        async def consumer() -> None:
+            async with queue.iterator() as queue_iterator:
+                async for message in queue_iterator:
+                    async with message.process():
+                        async with condition:
+                            data.append(message.body)
+                            messages.task_done()
+                            condition.notify()
 
-        count = 0
-        data = list()
+        async def application_stop_request() -> None:
+            async with condition:
+                await condition.wait_for(messages.full)
+            await messages.join()
+            await asyncio.sleep(1)
+            await connection.close()
 
-        async with queue.iterator() as queue_iterator:
-            async for message in queue_iterator:
-                async with message.process():
-                    count += 1
-                    data.append(message.body)
+        p = event_loop.create_task(publisher())
+        c = event_loop.create_task(consumer())
+        asr = event_loop.create_task(application_stop_request())
 
-                if count >= messages:
-                    break
+        data: List[bytes] = list()
 
-        assert data == list(map(lambda x: str(x).encode(), range(messages)))
+        await asyncio.gather(p, c, asr)
+
+        assert data == list(
+            map(lambda x: str(x).encode(), range(messages.maxsize))
+        )
 
     async def test_async_with_connection(
-        self, create_connection: Callable, connection, loop, declare_queue,
+        self, create_connection: Callable,
+        connection, event_loop, declare_queue,
     ):
         async with await create_connection() as connection:
 
@@ -1289,32 +1348,47 @@ class TestCaseAmqp(TestCaseAmqpBase):
                 channel=channel,
             )
 
-            messages = 100
+            condition = asyncio.Condition()
+            messages: asyncio.Queue[bytes] = asyncio.Queue(
+                100
+            )
 
             async def publisher():
                 channel1 = await self.create_channel(connection)
 
-                for i in range(messages):
+                for i in range(messages.maxsize):
+                    body = str(i).encode()
+                    await messages.put(body)
                     await channel1.default_exchange.publish(
-                        Message(body=str(i).encode()), routing_key=queue.name,
+                        Message(body=body), routing_key=queue.name,
                     )
 
-            loop.create_task(publisher())
-
-            count = 0
             data = list()
 
-            async with queue.iterator() as queue_iterator:
-                async for message in queue_iterator:
-                    async with message.process():
-                        count += 1
-                        data.append(message.body)
+            async def consume_loop():
+                async with queue.iterator() as queue_iterator:
+                    async for message in queue_iterator:
+                        async with message.process():
+                            async with condition:
+                                data.append(message.body)
+                                condition.notify()
+                                messages.task_done()
 
-                    if count >= messages:
-                        break
+            async def application_close_request():
+                async with condition:
+                    await condition.wait_for(messages.full)
+                await messages.join()
+                await asyncio.sleep(1)
+                await connection.close()
+
+            p = event_loop.create_task(publisher())
+            cl = event_loop.create_task(consume_loop())
+            acr = event_loop.create_task(application_close_request())
+
+            await asyncio.gather(p, cl, acr)
 
             assert data == list(
-                map(lambda x: str(x).encode(), range(messages)),
+                map(lambda x: str(x).encode(), range(messages.maxsize)),
             )
 
         assert channel.is_closed
@@ -1368,58 +1442,48 @@ class TestCaseAmqp(TestCaseAmqpBase):
             await q2.consume(print, exclusive=True)
 
     async def test_queue_iterator_close_was_called_twice(
-        self, create_connection: Callable, loop, declare_queue,
+        self, create_connection: Callable, event_loop, declare_queue,
     ):
-        future = loop.create_future()
         event = asyncio.Event()
 
         queue_name = get_random_name()
+        iterator: QueueIterator
 
         async def task_inner():
-            nonlocal future
             nonlocal event
+            nonlocal iterator
             nonlocal create_connection
 
-            try:
-                connection = await create_connection()
+            connection = await create_connection()
 
-                async with connection:
-                    channel = await self.create_channel(connection)
+            async with connection:
+                channel = await self.create_channel(connection)
 
-                    queue = await declare_queue(
-                        queue_name, channel=channel, cleanup=False,
-                    )
+                queue = await declare_queue(
+                    queue_name, channel=channel, cleanup=False,
+                )
 
-                    async with queue.iterator() as q:
-                        event.set()
+                async with queue.iterator() as iterator:
+                    event.set()
 
-                        async for message in q:
-                            with message.process():
-                                break
+                    async for message in iterator:
+                        async with message.process():
+                            pytest.fail("who sent this message?")
 
-            except asyncio.CancelledError as e:
-                future.set_exception(e)
-                raise
-
-        task = loop.create_task(task_inner())
+        task = event_loop.create_task(task_inner())
 
         await event.wait()
-        loop.call_soon(task.cancel)
-
-        with pytest.raises(asyncio.CancelledError):
-            await task
-
-        with pytest.raises(asyncio.CancelledError):
-            await future
+        await iterator.close()
+        await task
 
     async def test_queue_iterator_close_with_noack(
         self,
         create_connection: Callable,
-        loop,
+        event_loop,
         add_cleanup: Callable,
         declare_queue,
     ):
-        messages = []
+        messages: asyncio.Queue = asyncio.Queue()
         queue_name = get_random_name("test_queue")
         body = get_random_name("test_body").encode()
 
@@ -1440,7 +1504,7 @@ class TestCaseAmqp(TestCaseAmqpBase):
 
                 async with queue.iterator(no_ack=True) as q:
                     async for message in q:
-                        messages.append(message)
+                        await messages.put(message)
                         return
 
         async with await create_connection() as connection:
@@ -1455,16 +1519,72 @@ class TestCaseAmqp(TestCaseAmqpBase):
                     Message(body), routing_key=queue_name,
                 )
 
-                task = loop.create_task(task_inner())
+                task = event_loop.create_task(task_inner())
 
-                await task
+                message = await messages.get()
 
-                assert messages
-                assert messages[0].body == body
+                assert message
+                assert message.body == body
 
             finally:
+                await task
                 await queue.delete()
 
+    async def test_queue_iterator_throws_cancelled_error(
+        self,
+        create_connection: Callable,
+        event_loop,
+        add_cleanup: Callable,
+        declare_queue,
+    ):
+        event_loop.set_debug(True)
+        queue_name = get_random_name("test_queue")
+
+        connection = await create_connection()
+
+        async with connection:
+            channel = await self.create_channel(connection)
+
+            queue = await channel.declare_queue(
+                queue_name,
+            )
+
+            iterator = queue.iterator()
+            task = event_loop.create_task(iterator.__anext__())
+            done, pending = await asyncio.wait({task}, timeout=1)
+            assert not done
+            task.cancel()
+
+            with pytest.raises(asyncio.CancelledError):
+                await task
+
+    async def test_queue_iterator_throws_timeout_error(
+        self,
+        create_connection: Callable,
+        event_loop,
+        add_cleanup: Callable,
+        declare_queue,
+    ):
+        event_loop.set_debug(True)
+        queue_name = get_random_name("test_queue")
+
+        connection = await create_connection()
+
+        async with connection:
+            channel = await self.create_channel(connection)
+
+            queue = await channel.declare_queue(
+                queue_name,
+            )
+
+            iterator = queue.iterator(timeout=1)
+            task = event_loop.create_task(iterator.__anext__())
+            done, pending = await asyncio.wait({task}, timeout=5)
+            assert done
+
+            with pytest.raises(asyncio.TimeoutError):
+                await task
+
     async def test_passive_for_exchange(
         self, declare_exchange: Callable, connection, add_cleanup: Callable,
     ):
@@ -1561,15 +1681,152 @@ class TestCaseAmqp(TestCaseAmqpBase):
             await channel.set_qos(10)
 
     async def test_heartbeat_disabling(
-        self, loop, amqp_url: URL, connection_fabric,
+        self, event_loop, amqp_url: URL, connection_fabric,
     ):
         url = amqp_url.update_query(heartbeat=0)
         connection: AbstractConnection = await connection_fabric(url)
 
+        transport = connection.transport
+        assert transport
+        heartbeat = transport.connection.connection_tune.heartbeat
+
         async with connection:
-            assert (
-                connection.transport.connection.connection_tune.heartbeat == 0
-            )
+            assert heartbeat == 0
+
+    async def test_non_acked_messages_are_redelivered_to_queue(
+        self,
+        channel: aio_pika.Channel,
+        declare_queue: Callable,
+        declare_exchange: Callable,
+    ):
+        queue_name = get_random_name("test_connection")
+        routing_key = get_random_name()
+
+        exchange = await declare_exchange(
+            "direct", auto_delete=True, channel=channel,
+        )
+
+        queue: AbstractQueue = await declare_queue(
+            queue_name, auto_delete=False, channel=channel,
+        )
+
+        await queue.bind(exchange, routing_key)
+
+        # Publish 5 messages to queue
+        all_bodies = []
+        for _ in range(0, 5):
+            body = bytes(shortuuid.uuid(), "utf-8")
+            all_bodies.append(body)
+
+            assert await exchange.publish(Message(body), routing_key)
+
+        # Create a subscription but only process first message
+        async with queue.iterator() as queue_iterator:
+            first_message = await queue_iterator.__anext__()
+            async with first_message.process():
+                assert first_message.body == all_bodies[0]
+
+        # Confirm other messages are still in queue
+        for i in range(1, 5):
+            incoming_message = await queue.get(timeout=5)
+            await incoming_message.ack()
+
+            assert incoming_message.body == all_bodies[i]
+
+        # Check if the queue is now empty
+        assert await queue.get(fail=False, timeout=.5) is None
+
+        # Cleanup, delete the queue
+        await queue.delete()
+
+    async def test_regression_only_messages_cancelled_subscription_are_nacked(
+        self,
+        channel: aio_pika.Channel,
+        declare_queue: Callable,
+        declare_exchange: Callable,
+    ):
+        queue_name1 = get_random_name("test_queue")
+        queue_name2 = get_random_name("test_queue")
+        routing_key1 = get_random_name()
+        routing_key2 = get_random_name()
+
+        exchange = await declare_exchange(
+            "direct", auto_delete=True, channel=channel,
+        )
+
+        queue1: AbstractQueue = await declare_queue(
+            queue_name1, auto_delete=False, channel=channel,
+        )
+        queue2: AbstractQueue = await declare_queue(
+            queue_name2, auto_delete=False, channel=channel,
+        )
+
+        await queue1.bind(exchange, routing_key1)
+        await queue2.bind(exchange, routing_key2)
+
+        # Publish 5 messages to queue 1
+        all_bodies1 = []
+        for _ in range(0, 5):
+            body = bytes(shortuuid.uuid(), "utf-8")
+            all_bodies1.append(body)
+
+            assert await exchange.publish(Message(body), routing_key1)
+
+        # Publish 5 messages to queue 2
+        all_bodies2 = []
+        for _ in range(0, 5):
+            body = bytes(shortuuid.uuid(), "utf-8")
+            all_bodies2.append(body)
+
+            assert await exchange.publish(Message(body), routing_key2)
+
+        # Create a subscription to both queues but only process first message
+        queue_iterator1 = await queue1.iterator().__aenter__()
+        queue_iterator2 = await queue2.iterator().__aenter__()
+
+        first_message1 = await queue_iterator1.__anext__()
+        async with first_message1.process():
+            assert first_message1.body == all_bodies1[0]
+
+        first_message2 = await queue_iterator2.__anext__()
+        async with first_message2.process():
+            assert first_message2.body == all_bodies2[0]
+        #  The order of exit here is important.
+        #    Subscription to queue 1 is received first then to 2.
+        #    Therefore, the delivery tags of subscription to queue 2 will be
+        #    higher.
+        #    So first we cancel the subscription to 2, to test if we
+        #    accidentally also nacked the messages of queue 1. Then we cancel
+        #    subscription to queue 1 to test.
+
+        await queue_iterator2.__aexit__(None, None, None)
+        # To test if the wrong messages are nacked by stopping subscription to
+        # queue 2, we ack a message received from queue 1. If it was nacked,
+        # RabbitMQ will throw an exception.
+        second_message1 = await queue_iterator1.__anext__()
+        async with second_message1.process():
+            assert second_message1.body == all_bodies1[1]
+
+        await queue_iterator1.__aexit__(None, None, None)
+
+        # Confirm other messages are still in queue
+        for i in range(2, 5):
+            incoming_message = await queue1.get(timeout=5)
+            await incoming_message.ack()
+            assert incoming_message.body == all_bodies1[i]
+
+        for i in range(1, 5):
+            incoming_message = await queue2.get(timeout=5)
+            await incoming_message.ack()
+            assert incoming_message.body == all_bodies2[i]
+
+        # Check if the queue is now empty
+        assert await queue1.get(fail=False, timeout=.5) is None
+        assert await queue2.get(fail=False, timeout=.5) is None
+
+        # Cleanup, delete the queue
+        await queue1.delete()
+        await queue2.delete()
 
 
 class TestCaseAmqpNoConfirms(TestCaseAmqp):
@@ -1597,7 +1854,7 @@ class TestCaseAmqpWithConfirms(TestCaseAmqpBase):
         try:
             with pytest.raises(aio_pika.exceptions.ChannelPreconditionFailed):
                 msg = Message(bytes(shortuuid.uuid(), "utf-8"))
-                msg.delivery_mode = 8
+                msg.delivery_mode = 8       # type: ignore
 
                 await exchange.publish(msg, routing_key)
 
@@ -1608,12 +1865,12 @@ class TestCaseAmqpWithConfirms(TestCaseAmqpBase):
         finally:
             await exchange.delete()
 
-    async def test_basic_return(self, connection: aio_pika.connection, loop):
-        channel = await self.create_channel(
-            connection,
-        )  # type: aio_pika.Channel
+    async def test_basic_return(
+        self, connection: aio_pika.Connection, event_loop,
+    ):
+        channel = await self.create_channel(connection)
 
-        f = loop.create_future()
+        f = event_loop.create_future()
 
         def handler(channel, message: ReturnedMessage):
             f.set_result(message)
@@ -1632,13 +1889,11 @@ class TestCaseAmqpWithConfirms(TestCaseAmqpBase):
         assert returned.body == body
 
         # handler with exception
-        f = loop.create_future()
+        f = event_loop.create_future()
 
         await channel.close()
 
-        channel = await self.create_channel(
-            connection,
-        )  # type: aio_pika.Channel
+        channel = await self.create_channel(connection)
 
         def bad_handler(
             channel: aio_pika.abc.AbstractChannel,
