@@ -4,7 +4,7 @@ from functools import partial
 import pytest
 
 import aio_pika
-import tests
+from tests import test_amqp as amqp
 
 
 @pytest.fixture(
@@ -15,15 +15,15 @@ def connection_fabric(request):
 
 
 @pytest.fixture
-def create_connection(connection_fabric, loop, amqp_url):
+def create_connection(connection_fabric, event_loop, rabbitmq_container):
     ssl_context = ssl.create_default_context()
     ssl_context.check_hostname = False
     ssl_context.verify_mode = ssl.VerifyMode.CERT_NONE
 
     return partial(
         connection_fabric,
-        amqp_url.with_scheme("amqps").with_port(5671),
-        loop=loop,
+        rabbitmq_container.get_amqps_url(),
+        loop=event_loop,
         ssl_context=ssl_context,
     )
 
@@ -44,5 +44,5 @@ async def test_default_context(connection_fabric, amqp_url):
         )
 
 
-class TestCaseAMQPS(tests.test_amqp.TestCaseAmqp):
+class TestCaseAMQPS(amqp.TestCaseAmqp):
     pass
