@@ -1,5 +1,6 @@
 import gc
 import weakref
+from typing import AbstractSet
 
 import aio_pika
 
@@ -7,7 +8,7 @@ import aio_pika
 async def test_leak_unclosed_channel(create_connection):
     rabbitmq_connection = await create_connection()
 
-    weakset = weakref.WeakSet()
+    weakset: AbstractSet[aio_pika.abc.AbstractChannel] = weakref.WeakSet()
 
     async def f(rabbitmq_connection: aio_pika.Connection, weakset):
         weakset.add(await rabbitmq_connection.channel())
@@ -24,7 +25,7 @@ async def test_leak_unclosed_channel(create_connection):
 async def test_leak_closed_channel(create_connection):
     rabbitmq_connection = await create_connection()
 
-    weakset = weakref.WeakSet()
+    weakset: AbstractSet[aio_pika.abc.AbstractConnection] = weakref.WeakSet()
 
     async def f(rabbitmq_connection: aio_pika.Connection, weakset):
         async with rabbitmq_connection.channel() as channel:
